function [x, errors] = linesearchSQP(f, ...
                                     f_grad, ...
                                     c_eq, ...
                                     c_eq_Jacobi, ...
                                     c_ineq, ...
                                     lagrangianHessian, ...
                                     x, ...
                                     lambda, ...
                                     eta, ...
                                     tau, ...
                                     max_iterations, ...
                                     tolerance, ...
                                     x_solution)
    r = length(x);
    s = size(c_eq_Jacobi, 1);
    
    errors = zeros(max_iterations, 1);
    
    % Algoritm settings
    rho = .5;
    tau_alpha = .9 * tau;
    
    % Compute initial model.
    f_grad_x = f_grad(x);
    H_x = lagrangianHessian(x);
    c_eq_x = c_eq(x);
    KKT_matrix = [H_x         -c_eq_Jacobi';
                  c_eq_Jacobi spalloc(s, s, 0)];
    KKT_rhs = [-f_grad_x + c_eq_Jacobi' * lambda; -c_eq_x];
    
    sigma = 1;
    %mu = 1 + norm(lambda, Inf);
    
    %% SQP Algorithm
    for k = 1:max_iterations
        % Convergence test
        if norm(KKT_rhs) < tolerance
            fprintf('Converged after %d iterations with error %4.6g\n', k, norm(KKT_rhs));
            break
        end

        % Solve QR problem
        p_and_p_lambda = KKT_matrix \ KKT_rhs;
        p = p_and_p_lambda(1:r);
        p_lambda = p_and_p_lambda(r+1:end);
        

        % Compute mu_k
        mu_lower_lim = (dot(f_grad_x, p) + .5 * sigma * dot(p, H_x * p)) / ((1 - rho) * norm(c_eq_x, 1));
        if mu_lower_lim > 0;
            mu = 1.1 * mu_lower_lim;
        else
            mu = 1.1 * max(norm(lambda, Inf), norm(lambda + p_lambda, Inf));
        end

        % Compute step length
        alpha = backtracking(@phi, @phi_directional, f, f_grad, c_eq, x, 1, p, mu, eta, tau_alpha, c_ineq);
        
        % Update steps and model
        x = x + alpha * p;
        lambda = lambda + alpha * p_lambda;
        
        f_grad_x = f_grad(x);
        H_x = lagrangianHessian(x);
        c_eq_x = c_eq(x);
        KKT_matrix(1:r, 1:r) = H_x;
        KKT_rhs = [-f_grad_x + c_eq_Jacobi' * lambda; -c_eq_x];
        
        % Storing the errors
        errors(k) = norm(x_solution - x);
    end
    
    errors = errors(1:k);
    
    if k == max_iterations
        fprintf('Did not converge after %d iterations! Error: %4.6g\n', k, norm(KKT_rhs))
    end
end

%% Construct merit function and its directional derivative for backtracking.
function [y] = phi(f, c, x, mu)
    y = f(x) + mu * norm(c(x), 1);
end

function [y] = phi_directional(f_grad, c, x, p, mu)
    y = dot(f_grad(x), p) - mu * norm(c(x), 1);
end

%% Armijo backtracking
function [alpha] = backtracking(phi, phi_directional, f, f_grad, c, x, alpha, p, mu, eta, tau_alpha, c_ineq)
    phi_x = phi(f, c, x, mu);
    phi_dir_x_p = eta * phi_directional(f_grad, c, x, p, mu);
    
    while min(c_ineq(x + alpha * p)) <= 0
       alpha = .5 * alpha;
    end
    
    while phi(f, c, x + alpha * p, mu) > phi_x + alpha * phi_dir_x_p
        alpha = tau_alpha * alpha;
    end
end